@@ -1,23 +1,20 @@
-/* $Id: color.h,v 1.4 2006/07/10 23:25:37 ellson Exp $ $Revision: 1.4 $ */
+/* $Id: color.h,v 1.8 2011/01/25 16:30:48 ellson Exp $ $Revision: 1.8 $ */
 /* vim:set shiftwidth=4 ts=8: */
 
-/**********************************************************
-*      This software is part of the graphviz package      *
-*                http://www.graphviz.org/                 *
-*                                                         *
-*            Copyright (c) 1994-2004 AT&T Corp.           *
-*                and is licensed under the                *
-*            Common Public License, Version 1.0           *
-*                      by AT&T Corp.                      *
-*                                                         *
-*        Information and Software Systems Research        *
-*              AT&T Research, Florham Park NJ             *
-**********************************************************/
+/*************************************************************************
+ * Copyright (c) 2011 AT&T Intellectual Property 
+ * All rights reserved. This program and the accompanying materials
+ * are made available under the terms of the Eclipse Public License v1.0
+ * which accompanies this distribution, and is available at
+ * http://www.eclipse.org/legal/epl-v10.html
+ *
+ * Contributors: See CVS logs. Details at http://www.graphviz.org/
+ *************************************************************************/
 
 #ifndef GV_COLOR_H
 #define GV_COLOR_H
 
-#include "arith.h"
+/* #include "arith.h" */
 
 #ifdef __cplusplus
 extern "C" {
