@@ -1,18 +1,15 @@
-/* $Id: types.h,v 1.51 2006/06/18 03:21:20 ellson Exp $ $Revision: 1.51 $ */
+/* $Id: types.h,v 1.106 2011/02/15 16:27:53 erg Exp $ $Revision: 1.106 $ */
 /* vim:set shiftwidth=4 ts=8: */
 
-/**********************************************************
-*      This software is part of the graphviz package      *
-*                http://www.graphviz.org/                 *
-*                                                         *
-*            Copyright (c) 1994-2004 AT&T Corp.           *
-*                and is licensed under the                *
-*            Common Public License, Version 1.0           *
-*                      by AT&T Corp.                      *
-*                                                         *
-*        Information and Software Systems Research        *
-*              AT&T Research, Florham Park NJ             *
-**********************************************************/
+/*************************************************************************
+ * Copyright (c) 2011 AT&T Intellectual Property 
+ * All rights reserved. This program and the accompanying materials
+ * are made available under the terms of the Eclipse Public License v1.0
+ * which accompanies this distribution, and is available at
+ * http://www.eclipse.org/legal/epl-v10.html
+ *
+ * Contributors: See CVS logs. Details at http://www.graphviz.org/
+ *************************************************************************/
 
 #ifndef GV_TYPES_H
 #define GV_TYPES_H
@@ -21,13 +18,21 @@
 #include <assert.h>
 #include <signal.h>
 
-#include "cdt.h"
+typedef unsigned char boolean;
+#ifndef NOT
+#define NOT(v) (!(v))
+#endif
+#ifndef FALSE
+#define FALSE 0
+#endif
+#ifndef TRUE
+#define TRUE NOT(FALSE)
+#endif
+
 #include "geom.h"
-#include "pathplan.h"
-#include "color.h"
 #include "gvcext.h"
+#include "pathgeom.h"
 #include "textpara.h"
-#include "usershape.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -36,10 +41,20 @@ extern "C" {
     typedef int (*qsort_cmpf) (const void *, const void *);
     typedef int (*bsearch_cmpf) (const void *, const void *);
 
+#ifdef WITH_CGRAPH
+#include <cgraph.h>
+    typedef struct Agraph_s graph_t;
+    typedef struct Agnode_s node_t;
+    typedef struct Agedge_s edge_t;
+    typedef struct Agsym_s attrsym_t;
+#define TAIL_ID "tailport"
+#define HEAD_ID "headport"
+#else
     typedef struct Agraph_t graph_t;
     typedef struct Agnode_t node_t;
     typedef struct Agedge_t edge_t;
     typedef struct Agsym_t attrsym_t;
+#endif
 
     typedef struct htmllabel_t htmllabel_t;
 
@@ -50,52 +65,55 @@ extern "C" {
 	} a;
 	struct {
 	    node_t* n;
-	    box*    bp;
+	    boxf*    bp;
 	} s;
     } inside_t;
 
     typedef struct port {	/* internal edge endpoint specification */
-	point p;		/* aiming point relative to node center */
+	pointf p;		/* aiming point relative to node center */
 	double theta;		/* slope in radians */
-	box *bp;		/* if not null, points to bbox of 
+	boxf *bp;		/* if not null, points to bbox of 
 				 * rectangular area that is port target
 				 */
-	bool	defined;        /* if true, edge has port info at this end */
-	bool	constrained;    /* if true, constraints such as theta are set */
-	bool clip;           /* if true, clip end to node/port shape */
+	boolean	defined;        /* if true, edge has port info at this end */
+	boolean	constrained;    /* if true, constraints such as theta are set */
+	boolean clip;           /* if true, clip end to node/port shape */
+	boolean dyna;           /* if true, assign compass point dynamically */
 	unsigned char order;	/* for mincross */
 	unsigned char side;	/* if port is on perimeter of node, this
                                  * contains the bitwise OR of the sides (TOP,
                                  * BOTTOM, etc.) it is on. 
                                  */
+	char *name;		/* port name, if it was explicitly given, otherwise NULL */
     } port;
 
     typedef struct {
-	bool(*swapEnds) (edge_t * e);	/* Should head and tail be swapped? */
-	bool(*splineMerge) (node_t * n);	/* Is n a node in the middle of an edge? */
+	boolean(*swapEnds) (edge_t * e);	/* Should head and tail be swapped? */
+	boolean(*splineMerge) (node_t * n);	/* Is n a node in the middle of an edge? */
+	boolean ignoreSwap;                     /* Test for swapped edges if false */
+	boolean isOrtho;                        /* Orthogonal routing used */
     } splineInfo;
 
     typedef struct pathend_t {
-	box nb;			/* the node box */
-	point np;		/* node port */
+	boxf nb;			/* the node box */
+	pointf np;		/* node port */
 	int sidemask;
 	int boxn;
-	box boxes[20];
+	boxf boxes[20];
     } pathend_t;
 
     typedef struct path {	/* internal specification for an edge spline */
 	port start, end;
-	point *ulpp, *urpp, *llpp, *lrpp;	/* tangents of near splines */
 	int nbox;		/* number of subdivisions */
-	box *boxes;		/* rectangular regions of subdivision */
+	boxf *boxes;		/* rectangular regions of subdivision */
 	void *data;
     } path;
 
     typedef struct bezier {
-	point *list;
+	pointf *list;
 	int size;
 	int sflag, eflag;
-	point sp, ep;
+	pointf sp, ep;
     } bezier;
 
     typedef struct splines {
@@ -104,25 +122,15 @@ extern "C" {
 	boxf bb;
     } splines;
 
-/* fp variants */
-    typedef struct bezierf {
-	pointf *list;
-	int size;
-	int sflag, eflag;
-	pointf sp, ep;
-    } bezierf;
-
-    typedef struct splinesf {
-	bezierf *list;
-	int size;
-    } splinesf;
-
     typedef struct textlabel_t {
 	char *text, *fontname, *fontcolor;
+	int charset;
 	double fontsize;
-	pointf dimen;
-	point p;
-	pointf d;		/* delta from resizing */
+	pointf dimen; /* the diagonal size of the label (estimated by layout) */
+	pointf space; /* the diagonal size of the space for the label */
+		      /*   the rendered label is aligned in this box */
+		      /*   space does not include pad or margin */
+	pointf pos;   /* the center of the space for the label */
 	union {
 	    struct {
 		textpara_t *para;
@@ -130,8 +138,9 @@ extern "C" {
 	    } txt;
 	    htmllabel_t *html;
 	} u;
-	bool set;		/* true if position is set */
-	bool html;		/* true if html label */
+	char valign;  /* 't' 'c' 'b' */
+	boolean set;  /* true if position is set */
+	boolean html; /* true if html label */
     } textlabel_t;
 
     typedef struct polygon_t {	/* mutable shape information for a node */
@@ -169,72 +178,21 @@ extern "C" {
 	void (*initfn) (node_t *);	/* initializes shape from node u.shape_info structure */
 	void (*freefn) (node_t *);	/* frees  shape from node u.shape_info structure */
 	 port(*portfn) (node_t *, char *, char *);	/* finds aiming point and slope of port */
-	 bool(*insidefn) (inside_t * inside_context, pointf);	/* clips incident gvc->e spline on shape of gvc->n */
-	int (*pboxfn)(node_t* n, port* p, int side, box rv[], int *kptr); /* finds box path to reach port */
+	 boolean(*insidefn) (inside_t * inside_context, pointf);	/* clips incident gvc->e spline on shape of gvc->n */
+	int (*pboxfn)(node_t* n, port* p, int side, boxf rv[], int *kptr); /* finds box path to reach port */
 	void (*codefn) (GVJ_t * job, node_t * n);	/* emits graphics code for node */
     } shape_functions;
 
-    typedef enum { SH_UNSET, SH_POLY, SH_RECORD, SH_POINT, SH_EPSF,
-	    SH_USER } shape_kind;
+    typedef enum { SH_UNSET, SH_POLY, SH_RECORD, SH_POINT, SH_EPSF} shape_kind;
 
     typedef struct shape_desc {	/* read-only shape descriptor */
 	char *name;		/* as read from graph file */
 	shape_functions *fns;
 	polygon_t *polygon;	/* base polygon info */
-	bool usershape;
+	boolean usershape;
     } shape_desc;
 
-#ifdef WITH_CODEGENS
-    struct codegen_s {
-	void (*reset) (void);
-	void (*begin_job) (FILE * ofp, graph_t * g, char **lib, char *user,
-			   char *info[], point pages);
-	void (*end_job) (void);
-	void (*begin_graph) (GVC_t * gvc, graph_t * g, box bb, point pb);
-	void (*end_graph) (void);
-	void (*begin_page) (graph_t * g, point page, double scale, int rot,
-			    point offset);
-	void (*end_page) (void);
-	void (*begin_layer) (char *layerName, int n, int nLayers);
-	void (*end_layer) (void);
-	void (*begin_cluster) (graph_t * g);
-	void (*end_cluster) (void);
-	void (*begin_nodes) (void);
-	void (*end_nodes) (void);
-	void (*begin_edges) (void);
-	void (*end_edges) (void);
-	void (*begin_node) (node_t * n);
-	void (*end_node) (void);
-	void (*begin_edge) (edge_t * e);
-	void (*end_edge) (void);
-	void (*begin_context) (void);
-	void (*end_context) (void);
-	void (*begin_anchor) (char *href, char *tooltip, char *target);
-	void (*end_anchor) (void);
-	void (*set_font) (char *fontname, double fontsize);
-	void (*textpara) (point p, textpara_t * para);
-	void (*set_pencolor) (char *name);
-	void (*set_fillcolor) (char *name);
-	void (*set_style) (char **s);
-	void (*ellipse) (point p, int rx, int ry, int filled);
-	void (*polygon) (point * A, int n, int filled);
-	void (*beziercurve) (point * A, int n, int arrow_at_start,
-			     int arrow_at_end, int filled);
-	void (*polyline) (point * A, int n);
-	bool bezier_has_arrows;
-	void (*comment) (char *str);
-	void (*usershape) (usershape_t *us, boxf b, point * A, int sides, bool filled);
-    };
-
-    struct codegen_info_s {
-	codegen_t *cg;		/* discovered codegen */
-	char *name;		/* output format, null for sentinel */
-	int id;			/* id of output format */
-	void *info;		/* additional info provided by discovery routine */
-	/* Quartz uses this to store the Quicktime Component */
-    };
-
-#endif
+#include "usershape.h"		/* usershapes needed by gvc */
 
     typedef struct nodequeue {
 	node_t **store, **limit, **head, **tail;
@@ -252,8 +210,8 @@ extern "C" {
 	node_t **av;		/* allocated list of nodes in rank  */
 	int ht1, ht2;		/* height below/above centerline    */
 	int pht1, pht2;		/* as above, but only primitive nodes   */
-	bool candidate;	/* for transpose () */
-	bool valid;
+	boolean candidate;	/* for transpose () */
+	boolean valid;
 	int cache_nc;		/* caches number of crossings */
 	adjmatrix_t *flat;
     } rank_t;
@@ -266,19 +224,21 @@ extern "C" {
 	double scale;
 	double ratio;		/* set only if ratio_kind == R_VALUE */
 	double dpi;
-	point margin;
-	point page;
-	point  size;
-	bool filled;
-	bool landscape;
-	bool centered;
+	pointf margin;
+	pointf page;
+	pointf size;
+	boolean filled;
+	boolean landscape;
+	boolean centered;
 	ratio_t ratio_kind;
+	void* xdots;
+	char* id;
     } layout_t;
 
 /* for "record" shapes */
     typedef struct field_t {
-	point size;		/* its dimension */
-	box b;			/* its placement in node's coordinates */
+	pointf size;		/* its dimension */
+	boxf b;			/* its placement in node's coordinates */
 	int n_flds;
 	textlabel_t *lp;	/* n_flds == 0 */
 	struct field_t **fld;	/* n_flds > 0 */
@@ -307,15 +267,20 @@ extern "C" {
 #define alloc_elist(n,L)      do {L.size = 0; L.list = N_NEW(n + 1,edge_t*); } while (0)
 #define free_list(L)          do {if (L.list) free(L.list);} while (0)
 
+typedef enum {NATIVEFONTS,PSFONTS,SVGFONTS} fontname_kind;
+
     typedef struct Agraphinfo_t {
+#ifdef WITH_CGRAPH
+	Agrec_t hdr;
+#endif
 	/* to generate code */
 	layout_t *drawing;
 	textlabel_t *label;	/* if the cluster has a title */
-	box bb;			/* bounding box */
-	point border[4];	/* sizes of margins for graph labels */
+	boxf bb;			/* bounding box */
+	pointf border[4];	/* sizes of margins for graph labels */
 	unsigned char gui_state; /* Graph state for GUI ops */
 	unsigned char has_labels;
-	bool has_images;
+	boolean has_images;
 	unsigned char charset; /* input character set */
 	int rankdir;
 	int ht1, ht2;	/* below and above extremal ranks */
@@ -329,7 +294,8 @@ extern "C" {
 	node_t **neato_nlist;
 	int move;
 	double **dist, **spring, **sum_t, ***t;
-	int ndim;
+	unsigned short ndim;
+	unsigned short odim;
 #endif
 #ifndef NEATO_ONLY
 	/* to have subgraphs */
@@ -337,6 +303,12 @@ extern "C" {
 	graph_t **clust;	/* clusters are in clust[1..n_cluster] !!! */
 	node_t *nlist;
 	rank_t *rank;
+#ifdef WITH_CGRAPH
+	graph_t *parent;        /* containing cluster (not parent subgraph) */
+	int level;		/* cluster nesting level (not node level!) */
+	node_t	*minrep, *maxrep;	/* set leaders for min and max rank */
+#endif
+
 	/* fast graph node list */
 	nlist_t comp;
 	/* connected components */
@@ -346,25 +318,94 @@ extern "C" {
 	short minrank, maxrank;
 
 	/* various flags */
-	bool has_flat_edges;
+	boolean has_flat_edges;
+#ifdef WITH_CGRAPH
+	boolean has_sourcerank;
+	boolean has_sinkrank;
+#endif
 	unsigned char	showboxes;
-	bool cluster_was_collapsed;
+	boolean cluster_was_collapsed;
+	fontname_kind fontnames;		/* to override mangling in SVG */
 
 	int nodesep, ranksep;
 	node_t *ln, *rn;	/* left, right nodes of bounding box */
 
-
 	/* for clusters */
 	node_t *leader, **rankleader;
-	bool expanded;
+	boolean expanded;
 	char installed;
 	char set_type;
 	char label_pos;
-	bool exact_ranksep;
+	boolean exact_ranksep;
 #endif
 
     } Agraphinfo_t;
 
+#ifdef WITH_CGRAPH
+#define GD_parent(g) (((Agraphinfo_t*)AGDATA(g))->parent)
+#define GD_level(g) (((Agraphinfo_t*)AGDATA(g))->level)
+#define GD_drawing(g) (((Agraphinfo_t*)AGDATA(g))->drawing)
+#define GD_bb(g) (((Agraphinfo_t*)AGDATA(g))->bb)
+#define GD_gvc(g) (((Agraphinfo_t*)AGDATA(g))->gvc)
+#define GD_cleanup(g) (((Agraphinfo_t*)AGDATA(g))->cleanup)
+#define GD_dist(g) (((Agraphinfo_t*)AGDATA(g))->dist)
+#define GD_alg(g) (((Agraphinfo_t*)AGDATA(g))->alg)
+#define GD_border(g) (((Agraphinfo_t*)AGDATA(g))->border)
+#define GD_cl_cnt(g) (((Agraphinfo_t*)AGDATA(g))->cl_nt)
+#define GD_clust(g) (((Agraphinfo_t*)AGDATA(g))->clust)
+#define GD_cluster_was_collapsed(g) (((Agraphinfo_t*)AGDATA(g))->cluster_was_collapsed)
+#define GD_comp(g) (((Agraphinfo_t*)AGDATA(g))->comp)
+#define GD_exact_ranksep(g) (((Agraphinfo_t*)AGDATA(g))->exact_ranksep)
+#define GD_expanded(g) (((Agraphinfo_t*)AGDATA(g))->expanded)
+#define GD_flags(g) (((Agraphinfo_t*)AGDATA(g))->flags)
+#define GD_gui_state(g) (((Agraphinfo_t*)AGDATA(g))->gui_state)
+#define GD_charset(g) (((Agraphinfo_t*)AGDATA(g))->charset)
+#define GD_has_labels(g) (((Agraphinfo_t*)AGDATA(g))->has_labels)
+#define GD_has_images(g) (((Agraphinfo_t*)AGDATA(g))->has_images)
+#define GD_has_flat_edges(g) (((Agraphinfo_t*)AGDATA(g))->has_flat_edges)
+#define GD_has_sourcerank(g)	(((Agraphinfo_t*)AGDATA(g))->has_sourcerank)
+#define GD_has_sinkrank(g)	(((Agraphinfo_t*)AGDATA(g))->has_sinkrank)
+#define GD_ht1(g) (((Agraphinfo_t*)AGDATA(g))->ht1)
+#define GD_ht2(g) (((Agraphinfo_t*)AGDATA(g))->ht2)
+#define GD_inleaf(g) (((Agraphinfo_t*)AGDATA(g))->inleaf)
+#define GD_installed(g) (((Agraphinfo_t*)AGDATA(g))->installed)
+#define GD_label(g) (((Agraphinfo_t*)AGDATA(g))->label)
+#define GD_leader(g) (((Agraphinfo_t*)AGDATA(g))->leader)
+#define GD_rankdir2(g) (((Agraphinfo_t*)AGDATA(g))->rankdir)
+#define GD_rankdir(g) (((Agraphinfo_t*)AGDATA(g))->rankdir & 0x3)
+#define GD_flip(g) (GD_rankdir(g) & 1)
+#define GD_realrankdir(g) ((((Agraphinfo_t*)AGDATA(g))->rankdir) >> 2)
+#define GD_realflip(g) (GD_realrankdir(g) & 1)
+#define GD_ln(g) (((Agraphinfo_t*)AGDATA(g))->ln)
+#define GD_maxrank(g) (((Agraphinfo_t*)AGDATA(g))->maxrank)
+#define GD_maxset(g) (((Agraphinfo_t*)AGDATA(g))->maxset)
+#define GD_minrank(g) (((Agraphinfo_t*)AGDATA(g))->minrank)
+#define GD_minset(g) (((Agraphinfo_t*)AGDATA(g))->minset)
+#define GD_minrep(g) (((Agraphinfo_t*)AGDATA(g))->minrep)
+#define GD_maxrep(g) (((Agraphinfo_t*)AGDATA(g))->maxrep)
+#define GD_move(g) (((Agraphinfo_t*)AGDATA(g))->move)
+#define GD_n_cluster(g) (((Agraphinfo_t*)AGDATA(g))->n_cluster)
+#define GD_n_nodes(g) (((Agraphinfo_t*)AGDATA(g))->n_nodes)
+#define GD_ndim(g) (((Agraphinfo_t*)AGDATA(g))->ndim)
+#define GD_odim(g) (((Agraphinfo_t*)AGDATA(g))->odim)
+#define GD_neato_nlist(g) (((Agraphinfo_t*)AGDATA(g))->neato_nlist)
+#define GD_nlist(g) (((Agraphinfo_t*)AGDATA(g))->nlist)
+#define GD_nodesep(g) (((Agraphinfo_t*)AGDATA(g))->nodesep)
+#define GD_outleaf(g) (((Agraphinfo_t*)AGDATA(g))->outleaf)
+#define GD_rank(g) (((Agraphinfo_t*)AGDATA(g))->rank)
+#define GD_rankleader(g) (((Agraphinfo_t*)AGDATA(g))->rankleader)
+#define GD_ranksep(g) (((Agraphinfo_t*)AGDATA(g))->ranksep)
+#define GD_rn(g) (((Agraphinfo_t*)AGDATA(g))->rn)
+#define GD_set_type(g) (((Agraphinfo_t*)AGDATA(g))->set_type)
+#define GD_label_pos(g) (((Agraphinfo_t*)AGDATA(g))->label_pos)
+#define GD_showboxes(g) (((Agraphinfo_t*)AGDATA(g))->showboxes)
+#define GD_fontnames(g) (((Agraphinfo_t*)AGDATA(g))->fontnames)
+#define GD_spring(g) (((Agraphinfo_t*)AGDATA(g))->spring)
+#define GD_sum_t(g) (((Agraphinfo_t*)AGDATA(g))->sum_t)
+#define GD_t(g) (((Agraphinfo_t*)AGDATA(g))->t)
+
+#else
+
 #define GD_alg(g) (g)->u.alg
 #define GD_bb(g) (g)->u.bb
 #define GD_border(g) (g)->u.border
@@ -403,6 +444,7 @@ extern "C" {
 #define GD_n_cluster(g) (g)->u.n_cluster
 #define GD_n_nodes(g) (g)->u.n_nodes
 #define GD_ndim(g) (g)->u.ndim
+#define GD_odim(g) (g)->u.odim
 #define GD_neato_nlist(g) (g)->u.neato_nlist
 #define GD_nlist(g) (g)->u.nlist
 #define GD_nodesep(g) (g)->u.nodesep
@@ -414,32 +456,41 @@ extern "C" {
 #define GD_set_type(g) (g)->u.set_type
 #define GD_label_pos(g) (g)->u.label_pos
 #define GD_showboxes(g) (g)->u.showboxes
+#define GD_fontnames(g) (g)->u.fontnames
 #define GD_spring(g) (g)->u.spring
 #define GD_sum_t(g) (g)->u.sum_t
 #define GD_t(g) (g)->u.t
+#endif
 
     typedef struct Agnodeinfo_t {
+#ifdef WITH_CGRAPH
+	Agrec_t hdr;
+#endif
 	shape_desc *shape;
 	void *shape_info;
-	point coord;
-	double width, height;
+	pointf coord;
+	double width, height;  /* inches */
 	boxf bb;
-	int ht, lw, rw;
+	double ht, lw, rw;
 	textlabel_t *label;
+	textlabel_t *xlabel;
 	void *alg;
 	char state;
 	unsigned char gui_state; /* Node state for GUI ops */
-	bool clustnode;
+	boolean clustnode;
 
 #ifndef DOT_ONLY
 	unsigned char pinned;
-	short xsize, ysize;
 	int id, heapindex, hops;
 	double *pos, dist;
 #endif
 #ifndef NEATO_ONLY
 	unsigned char showboxes;
-	bool  has_port;
+	boolean  has_port;
+#ifdef WITH_CGRAPH
+	node_t* rep;
+	node_t *set;
+#endif
 
 	/* fast graph */
 	char node_type, mark, onstack;
@@ -469,12 +520,72 @@ extern "C" {
 
     } Agnodeinfo_t;
 
+#ifdef WITH_CGRAPH
+#define ND_id(n) (((Agnodeinfo_t*)AGDATA(n))->id)
+#define ND_alg(n) (((Agnodeinfo_t*)AGDATA(n))->alg)
+#define ND_UF_parent(n) (((Agnodeinfo_t*)AGDATA(n))->UF_parent)
+#define ND_set(n) (((Agnodeinfo_t*)AGDATA(n))->set)
+#define ND_UF_size(n) (((Agnodeinfo_t*)AGDATA(n))->UF_size)
+#define ND_bb(n) (((Agnodeinfo_t*)AGDATA(n))->bb)
+#define ND_clust(n) (((Agnodeinfo_t*)AGDATA(n))->clust)
+#define ND_coord(n) (((Agnodeinfo_t*)AGDATA(n))->coord)
+#define ND_dist(n) (((Agnodeinfo_t*)AGDATA(n))->dist)
+#define ND_flat_in(n) (((Agnodeinfo_t*)AGDATA(n))->flat_in)
+#define ND_flat_out(n) (((Agnodeinfo_t*)AGDATA(n))->flat_out)
+#define ND_gui_state(n) (((Agnodeinfo_t*)AGDATA(n))->gui_state)
+#define ND_has_port(n) (((Agnodeinfo_t*)AGDATA(n))->has_port)
+#define ND_rep(n) (((Agnodeinfo_t*)AGDATA(n))->rep)
+#define ND_heapindex(n) (((Agnodeinfo_t*)AGDATA(n))->heapindex)
+#define ND_height(n) (((Agnodeinfo_t*)AGDATA(n))->height)
+#define ND_hops(n) (((Agnodeinfo_t*)AGDATA(n))->hops)
+#define ND_ht(n) (((Agnodeinfo_t*)AGDATA(n))->ht)
+#define ND_in(n) (((Agnodeinfo_t*)AGDATA(n))->in)
+#define ND_inleaf(n) (((Agnodeinfo_t*)AGDATA(n))->inleaf)
+#define ND_label(n) (((Agnodeinfo_t*)AGDATA(n))->label)
+#define ND_xlabel(n) (((Agnodeinfo_t*)AGDATA(n))->xlabel)
+#define ND_lim(n) (((Agnodeinfo_t*)AGDATA(n))->lim)
+#define ND_low(n) (((Agnodeinfo_t*)AGDATA(n))->low)
+#define ND_lw(n) (((Agnodeinfo_t*)AGDATA(n))->lw)
+#define ND_mark(n) (((Agnodeinfo_t*)AGDATA(n))->mark)
+#define ND_mval(n) (((Agnodeinfo_t*)AGDATA(n))->mval)
+#define ND_n_cluster(n) (((Agnodeinfo_t*)AGDATA(n))->n_cluster)
+#define ND_next(n) (((Agnodeinfo_t*)AGDATA(n))->next)
+#define ND_node_type(n) (((Agnodeinfo_t*)AGDATA(n))->node_type)
+#define ND_onstack(n) (((Agnodeinfo_t*)AGDATA(n))->onstack)
+#define ND_order(n) (((Agnodeinfo_t*)AGDATA(n))->order)
+#define ND_other(n) (((Agnodeinfo_t*)AGDATA(n))->other)
+#define ND_out(n) (((Agnodeinfo_t*)AGDATA(n))->out)
+#define ND_outleaf(n) (((Agnodeinfo_t*)AGDATA(n))->outleaf)
+#define ND_par(n) (((Agnodeinfo_t*)AGDATA(n))->par)
+#define ND_pinned(n) (((Agnodeinfo_t*)AGDATA(n))->pinned)
+#define ND_pos(n) (((Agnodeinfo_t*)AGDATA(n))->pos)
+#define ND_prev(n) (((Agnodeinfo_t*)AGDATA(n))->prev)
+#define ND_priority(n) (((Agnodeinfo_t*)AGDATA(n))->priority)
+#define ND_rank(n) (((Agnodeinfo_t*)AGDATA(n))->rank)
+#define ND_ranktype(n) (((Agnodeinfo_t*)AGDATA(n))->ranktype)
+#define ND_rw(n) (((Agnodeinfo_t*)AGDATA(n))->rw)
+#define ND_save_in(n) (((Agnodeinfo_t*)AGDATA(n))->save_in)
+#define ND_save_out(n) (((Agnodeinfo_t*)AGDATA(n))->save_out)
+#define ND_shape(n) (((Agnodeinfo_t*)AGDATA(n))->shape)
+#define ND_shape_info(n) (((Agnodeinfo_t*)AGDATA(n))->shape_info)
+#define ND_showboxes(n) (((Agnodeinfo_t*)AGDATA(n))->showboxes)
+#define ND_state(n) (((Agnodeinfo_t*)AGDATA(n))->state)
+#define ND_clustnode(n) (((Agnodeinfo_t*)AGDATA(n))->clustnode)
+#define ND_tree_in(n) (((Agnodeinfo_t*)AGDATA(n))->tree_in)
+#define ND_tree_out(n) (((Agnodeinfo_t*)AGDATA(n))->tree_out)
+#define ND_weight_class(n) (((Agnodeinfo_t*)AGDATA(n))->weight_class)
+#define ND_width(n) (((Agnodeinfo_t*)AGDATA(n))->width)
+#define ND_xsize(n) (ND_lw(n)+ND_rw(n))
+#define ND_ysize(n) (ND_ht(n))
+
+#else
+
 #define ND_UF_parent(n) (n)->u.UF_parent
 #define ND_UF_size(n) (n)->u.UF_size
 #define ND_alg(n) (n)->u.alg
 #define ND_bb(n) (n)->u.bb
 #define ND_clust(n) (n)->u.clust
-#define ND_coord_i(n) (n)->u.coord
+#define ND_coord(n) (n)->u.coord
 #define ND_dist(n) (n)->u.dist
 #define ND_flat_in(n) (n)->u.flat_in
 #define ND_flat_out(n) (n)->u.flat_out
@@ -483,14 +594,15 @@ extern "C" {
 #define ND_heapindex(n) (n)->u.heapindex
 #define ND_height(n) (n)->u.height
 #define ND_hops(n) (n)->u.hops
-#define ND_ht_i(n) (n)->u.ht
+#define ND_ht(n) (n)->u.ht
 #define ND_id(n) (n)->u.id
 #define ND_in(n) (n)->u.in
 #define ND_inleaf(n) (n)->u.inleaf
 #define ND_label(n) (n)->u.label
+#define ND_xlabel(n) (n)->u.xlabel
 #define ND_lim(n) (n)->u.lim
 #define ND_low(n) (n)->u.low
-#define ND_lw_i(n) (n)->u.lw
+#define ND_lw(n) (n)->u.lw
 #define ND_mark(n) (n)->u.mark
 #define ND_mval(n) (n)->u.mval
 #define ND_n_cluster(n) (n)->u.n_cluster
@@ -508,7 +620,7 @@ extern "C" {
 #define ND_priority(n) (n)->u.priority
 #define ND_rank(n) (n)->u.rank
 #define ND_ranktype(n) (n)->u.ranktype
-#define ND_rw_i(n) (n)->u.rw
+#define ND_rw(n) (n)->u.rw
 #define ND_save_in(n) (n)->u.save_in
 #define ND_save_out(n) (n)->u.save_out
 #define ND_shape(n) (n)->u.shape
@@ -520,13 +632,17 @@ extern "C" {
 #define ND_tree_out(n) (n)->u.tree_out
 #define ND_weight_class(n) (n)->u.weight_class
 #define ND_width(n) (n)->u.width
-#define ND_xsize(n) (n)->u.xsize
-#define ND_ysize(n) (n)->u.ysize
+#define ND_xsize(n) (ND_lw(n)+ND_rw(n))
+#define ND_ysize(n) (ND_ht(n))
+#endif
 
     typedef struct Agedgeinfo_t {
+#ifdef WITH_CGRAPH
+	Agrec_t hdr;
+#endif
 	splines *spl;
-	port tail_port, head_port;	/* might be used someday */
-	textlabel_t *label, *head_label, *tail_label;
+	port tail_port, head_port;
+	textlabel_t *label, *head_label, *tail_label, *xlabel;
 	char edge_type;
 	char adjacent;          /* true for flat edge with adjacent nodes */
 	char label_ontop;
@@ -541,17 +657,45 @@ extern "C" {
 #endif
 #ifndef NEATO_ONLY
 	unsigned char showboxes;
-	bool conc_opp_flag;
+	boolean conc_opp_flag;
 	short xpenalty;
-	int weight;
+	float weight;
 	int cutvalue, tree_index;
 	short count;
 	unsigned short minlen;
 	edge_t *to_virt;
 #endif
-
     } Agedgeinfo_t;
 
+#ifdef WITH_CGRAPH
+#define ED_alg(e) (((Agedgeinfo_t*)AGDATA(e))->alg)
+#define ED_conc_opp_flag(e) (((Agedgeinfo_t*)AGDATA(e))->conc_opp_flag)
+#define ED_count(e) (((Agedgeinfo_t*)AGDATA(e))->count)
+#define ED_cutvalue(e) (((Agedgeinfo_t*)AGDATA(e))->cutvalue)
+#define ED_edge_type(e) (((Agedgeinfo_t*)AGDATA(e))->edge_type)
+#define ED_adjacent(e) (((Agedgeinfo_t*)AGDATA(e))->adjacent)
+#define ED_factor(e) (((Agedgeinfo_t*)AGDATA(e))->factor)
+#define ED_gui_state(e) (((Agedgeinfo_t*)AGDATA(e))->gui_state)
+#define ED_head_label(e) (((Agedgeinfo_t*)AGDATA(e))->head_label)
+#define ED_head_port(e) (((Agedgeinfo_t*)AGDATA(e))->head_port)
+#define ED_label(e) (((Agedgeinfo_t*)AGDATA(e))->label)
+#define ED_xlabel(e) (((Agedgeinfo_t*)AGDATA(e))->xlabel)
+#define ED_label_ontop(e) (((Agedgeinfo_t*)AGDATA(e))->label_ontop)
+#define ED_minlen(e) (((Agedgeinfo_t*)AGDATA(e))->minlen)
+#define ED_path(e) (((Agedgeinfo_t*)AGDATA(e))->path)
+#define ED_showboxes(e) (((Agedgeinfo_t*)AGDATA(e))->showboxes)
+#define ED_spl(e) (((Agedgeinfo_t*)AGDATA(e))->spl)
+#define ED_tail_label(e) (((Agedgeinfo_t*)AGDATA(e))->tail_label)
+#define ED_tail_port(e) (((Agedgeinfo_t*)AGDATA(e))->tail_port)
+#define ED_to_orig(e) (((Agedgeinfo_t*)AGDATA(e))->to_orig)
+#define ED_to_virt(e) (((Agedgeinfo_t*)AGDATA(e))->to_virt)
+#define ED_tree_index(e) (((Agedgeinfo_t*)AGDATA(e))->tree_index)
+#define ED_xpenalty(e) (((Agedgeinfo_t*)AGDATA(e))->xpenalty)
+#define ED_dist(e) (((Agedgeinfo_t*)AGDATA(e))->dist)
+#define ED_weight(e) (((Agedgeinfo_t*)AGDATA(e))->weight)
+
+#else
+
 #define ED_alg(e) (e)->u.alg
 #define ED_conc_opp_flag(e) (e)->u.conc_opp_flag
 #define ED_count(e) (e)->u.count
@@ -564,6 +708,7 @@ extern "C" {
 #define ED_head_label(e) (e)->u.head_label
 #define ED_head_port(e) (e)->u.head_port
 #define ED_label(e) (e)->u.label
+#define ED_xlabel(e) (e)->u.xlabel
 #define ED_label_ontop(e) (e)->u.label_ontop
 #define ED_minlen(e) (e)->u.minlen
 #define ED_path(e) (e)->u.path
@@ -576,6 +721,33 @@ extern "C" {
 #define ED_tree_index(e) (e)->u.tree_index
 #define ED_weight(e) (e)->u.weight
 #define ED_xpenalty(e) (e)->u.xpenalty
+#endif
+
+#ifdef WITH_CGRAPH
+#include "cgraph.h"
+#define SET_RANKDIR(g,rd) (GD_rankdir2(g) = rd)
+#define agfindedge(g,t,h) (agedge(g,t,h,NULL,0))
+#define agfindnode(g,n) (agnode(g,n,0))
+#define agfindgraphattr(g,a) (agattr(g,AGRAPH,a,NULL))
+#define agfindnodeattr(g,a) (agattr(g,AGNODE,a,NULL))
+#define agfindedgeattr(g,a) (agattr(g,AGEDGE,a,NULL))
+#else
+#include "graph.h"
+#define SET_RANKDIR(g,rd) ((g)->u.rankdir = (rd))
+#define agnameof(x) ((x)->name)
+    /* warning, agraphof doesn't work for edges */
+#define agraphof(n) ((n)->graph)
+#define agroot(g) ((g)->root)
+#define aghead(e) ((e)->head)
+#define agtail(e) ((e)->tail)
+#define agisdirected(g) ((g)->kind & AGFLAG_DIRECTED)
+#define AGID(x) ((x)->id)
+#define agfindgraphattr(g,a) agfindattr((g)->root,a)
+#define agfindnodeattr(g,a) agfindattr((g)->proto->n,a)
+#define agfindedgeattr(g,a) agfindattr((g)->proto->e,a)
+#define agcanonStr(s) agcanonical(s)
+#endif
+
 
     typedef struct {
         int useGrid;            /* use grid for speed up */
